# long2short/chunker.py
from typing import List, Tuple, Optional
from .tokenizer import Tokenizer


class TextChunker:
    """Handles splitting text into manageable chunks."""

    def __init__(self, tokenizer: Tokenizer):
        self.tokenizer = tokenizer

    def chunk_text(
        self,
        text: str,
        max_tokens: int,
        delimiter: str = ".",
        header: Optional[str] = None,
        add_ellipsis: bool = True,
    ) -> Tuple[List[str], List[List[int]], int]:
        """
        Split text into chunks based on token limit.

        Args:
            text: Text to split into chunks
            max_tokens: Maximum tokens per chunk
            delimiter: Delimiter to split text on
            header: Optional header to prepend to each chunk
            add_ellipsis: Whether to add ellipsis for overflow chunks

        Returns:
            Tuple of (chunks, chunk_indices, dropped_chunk_count)
        """
        chunks = text.split(delimiter)
        combined_chunks, indices, dropped = self._combine_chunks(
            chunks=chunks,
            max_tokens=max_tokens,
            chunk_delimiter=delimiter,
            header=header,
            add_ellipsis=add_ellipsis,
        )
        if dropped > 0:
            print(f"warning: {dropped} chunks were dropped due to overflow")

        # Add back delimiters
        combined_chunks = [f"{chunk}{delimiter}" for chunk in combined_chunks]
        return combined_chunks, indices, dropped

    def _combine_chunks(
        self,
        chunks: List[str],
        max_tokens: int,
        chunk_delimiter: str = "\n\n",
        header: Optional[str] = None,
        add_ellipsis: bool = True,
    ) -> Tuple[List[str], List[List[int]], int]:
        """
        Combine text chunks while respecting token limits.

        Args:
            chunks: List of text chunks to combine
            max_tokens: Maximum tokens per combined chunk
            chunk_delimiter: Delimiter to join chunks with
            header: Optional header to prepend to each chunk
            add_ellipsis: Whether to add ellipsis for overflow chunks
        """
        dropped_chunk_count = 0
        output = []
        output_indices = []
        candidate = [] if header is None else [header]

        candidate_indices = []
        for chunk_i, chunk in enumerate(chunks):
            chunk_with_header = [chunk] if header is None else [header, chunk]

            # Check if single chunk exceeds token limit
            if (len(self.tokenizer.tokenize(chunk_delimiter.join(chunk_with_header))) > max_tokens):
                print(f"Warning: chunk overflow")
                if add_ellipsis and len(self.tokenizer.tokenize(chunk_delimiter.join(candidate + ["..."]))) <= max_tokens:
                    candidate.append("...")
                    dropped_chunk_count += 1
                continue  # this case would break downstream assumptions

            # estimate token count with the current chunk added
            extended_candidate_token_count = len(self.tokenizer.tokenize(chunk_delimiter.join(candidate + [chunk])))
            # If the token count exceeds max_tokens, add the current candidate to output and start a new candidate
            if extended_candidate_token_count > max_tokens:
                output.append(chunk_delimiter.join(candidate))
                output_indices.append(candidate_indices)
                candidate = chunk_with_header  # re-initialize candidate
                candidate_indices = [chunk_i]
            # otherwise keep extending the candidate
            else:
                candidate.append(chunk)
                candidate_indices.append(chunk_i)
        # add the remaining candidate to output if it's not empty
        if (header is not None and len(candidate) > 1) or (header is None and len(candidate) > 0):
            output.append(chunk_delimiter.join(candidate))
            output_indices.append(candidate_indices)
        return output, output_indices, dropped_chunk_count


# long2short/__init__.py
from .models import LLMProvider, OpenAIProvider, AnthropicProvider
from .tokenizer import Tokenizer
from .chunker import TextChunker
from .long2short import Long2Short

__version__ = "0.1.0"

__all__ = [
    "LLMProvider",
    "OpenAIProvider",
    "AnthropicProvider",
    "Tokenizer",
    "TextChunker",
    "Long2Short",
]


# long2short/long2short.py
from typing import Optional
from tqdm import tqdm

from .models import LLMProvider
from .tokenizer import Tokenizer
from .chunker import TextChunker

class Long2Short:
    """Main class for text summarization."""
    
    def __init__(self, 
        llm_provider: LLMProvider,
        tokenizer: Optional[Tokenizer] = None):
        """
        Initialize summarizer with LLM provider and tokenizer.
        
        Args:
            llm_provider: Provider for language model completions
            tokenizer: Optional tokenizer (will create default if None)
        """
        self.llm_provider = llm_provider
        self.tokenizer = tokenizer or Tokenizer()
        self.chunker = TextChunker(self.tokenizer)
    
    def summarize(self,
        text: str,
        detail: float = 0,
        additional_instructions: Optional[str] = None,
        minimum_chunk_size: int = 500,
        chunk_delimiter: str = ".",
        header: Optional[str] = None,
        summarize_recursively: bool = False,
        verbose: bool = False,
        **kwargs) -> str:
        """
        Summarize text with controllable level of detail.
        
        Args:
            text: Input text to summarize
            detail: Level of detail (0 to 1), higher means more detailed summary
            additional_instructions: Extra instructions for the LLM
            minimum_chunk_size: Minimum size for text chunks
            chunk_delimiter: Delimiter for splitting text into chunks
            header: Optional header to prepend to each chunk
            summarize_recursively: Whether to use previous summaries as context
            verbose: Print detailed information about the process
            **kwargs: Additional arguments passed to the LLM provider
        
        Returns:
            Summarized text
        """
        assert 0 <= detail <= 1, "Detail must be between 0 and 1"

        # interpolate the number of chunks based to get specified level of detail
        max_chunks = len(self.chunker.chunk_text(text, minimum_chunk_size, chunk_delimiter))
        min_chunks = 1
        num_chunks = int(min_chunks + detail * (max_chunks - min_chunks))
        
        # Calculate number of chunks based on detail level
        document_length = len(self.tokenizer.tokenize(text))
        chunk_size = max(minimum_chunk_size, document_length // num_chunks)
        text_chunks, _, _ = self.chunker.chunk_text(text, chunk_size, chunk_delimiter)

        if verbose:
            print(f"Splitting the text into {len(text_chunks)} chunks to be summarized.")
            print(f"Chunk lengths are {[self.tokenizer.count_tokens(x) for x in text_chunks]}")                               
        
        # Build system message
        system_message = "Rewrite this text in summarized form."
        if additional_instructions is not None:
            system_message += f"\n\n{additional_instructions}"
        
        # Generate summaries
        accumulated_summaries = []
        for chunk in tqdm(text_chunks):
            if summarize_recursively and accumulated_summaries:
                accumulated_summaries_string = "\n\n".join(accumulated_summaries)
                prompt = f"Previous summaries:\n\n{accumulated_summaries_string}\n\nText to summarize next:\n\n{chunk}"
            else:
                prompt = chunk

            messages = [
                {"role": "system", "content": system_message},
                {"role": "user", "content": prompt}
            ]

            summary = self.llm_provider.generate_completion(messages, **kwargs)
            accumulated_summaries.append(summary)
        
        return "\n\n".join(accumulated_summaries)


# long2short/tokenizer.py
from typing import List
import tiktoken

class Tokenizer:
    """Handles text tokenization for different models."""
    
    def __init__(self, model: str = "gpt-4-turbo"):
        self.encoding = tiktoken.encoding_for_model(model)
    
    def tokenize(self, text: str) -> List[int]:
        """Convert text to tokens."""
        return self.encoding.encode(text)
    
    def detokenize(self, tokens: List[int]) -> str:
        """Convert tokens back to text."""
        return self.encoding.decode(tokens)
    
    def count_tokens(self, text: str) -> int:
        """Count the number of tokens in a text."""
        return len(self.tokenize(text))


# long2short/models.py
from abc import ABC, abstractmethod
from typing import List, Dict
import os

from openai import OpenAI
from anthropic import Anthropic

class LLMProvider(ABC):
    """Abstract base class for LLM providers."""
    
    @abstractmethod
    def generate_completion(self, prompt: str, **kwargs) -> str:
        """Generate completion from the LLM provider."""
        pass

class OpenAIProvider(LLMProvider):
    """OpenAI API provider implementation."""
    
    def __init__(self, api_key: str = None, model: str = "gpt-4-turbo", base_url: str = "https://api.openai.com/v1"):
        self.client = OpenAI(api_key=api_key or os.getenv("OPENAI_API_KEY"), base_url=base_url or os.getenv("OPENAI_API_BASE"))
        self.model = model
        self.base_url = base_url
    
    def generate_completion(self, messages: List[Dict[str, str]], **kwargs) -> str:        
        response = self.client.chat.completions.create(
            model=self.model,
            messages=messages,
            temperature=kwargs.get('temperature', 0),
        )
        return response.choices[0].message.content

class AnthropicProvider(LLMProvider):
    """Anthropic Claude API provider implementation."""
    
    def __init__(self, api_key: str = None, model: str = "claude-3-opus-20240229"):
        self.client = Anthropic(api_key=api_key or os.getenv("ANTHROPIC_API_KEY"))
        self.model = model
    
    def generate_completion(self, prompt: str, **kwargs) -> str:
        response = self.client.messages.create(
            model=self.model,
            messages=[{"role": "user", "content": f"Summarize the following text:\n\n{prompt}"}],
            temperature=kwargs.get('temperature', 0),
        )
        return response.content[0].text